/**
 * Copyright (C) 2007-2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <glib/gstdio.h>

#include <libsoup/soup.h>

#include "main.h"
#include "wk-slideshow-item.h"
#include "wk-settings.h"
#include "wk-daemon.h"
#include "wk-browser.h"
#include "wk-events.h"
#include "utils.h"

static void _daemon_select_playlist_loop (WkDaemon *d);

static gboolean
_daemon_browser_show_uri (WkDaemon *d, const gchar *uri)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("showing item: %s", uri);

    return wk_browser_show (browser, WK_SLIDESHOW_ITEM_TYPE_WEB, uri, 0);
}

static gboolean
_daemon_browser_show (WkDaemon *d, WkSlideshowItem *item)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("showing item: %s", item->uri);

    return wk_browser_show (browser, item->type, item->uri, item->timeout);
}

static gboolean
_daemon_show_next_item (gpointer data)
{
    WkSlideshowItem *item;
    WkDaemon *d = (WkDaemon *) data;

    DEBUG ("start");

    if (d->loop && ((d->current = g_slist_next (d->current)) == NULL))
    {
        DEBUG ("end of playlist loop");
        _daemon_select_playlist_loop (d);
    }

    if (! d->current)
    {
        d->current = d->loop->data;
    }

    item = (WkSlideshowItem *) d->current->data;

    /* if href attribute is true, this is only data used by other items. We
       don't show it */
    if (item->href)
    {
        DEBUG ("not showing href resource: %s", item->uri);
        _daemon_show_next_item (data);
        return FALSE;
    }

    /* Deal with idle. */
    if (item->type == WK_SLIDESHOW_ITEM_TYPE_IDLE)
    {
        if (d->status != WK_DAEMON_STATUS_IDLE)
        {
            DEBUG ("going to idle");
            g_spawn_command_line_async ("xset dpms force off", NULL);
            d->status = WK_DAEMON_STATUS_IDLE;
        }
        d->timeout_id = g_timeout_add_seconds (item->timeout,
                                               _daemon_show_next_item,
                                               d);
        return FALSE;
    }

    if (d->status == WK_DAEMON_STATUS_IDLE)
    {
        DEBUG ("exiting idle");
        d->status = WK_DAEMON_STATUS_SLIDESHOW;
        g_spawn_command_line_async ("xset dpms force on", NULL);
    }

    if (!_daemon_browser_show (d, item))
    {
        _daemon_show_next_item (d);
        return FALSE;
    }

    if (item->type != WK_SLIDESHOW_ITEM_TYPE_VIDEO)
    {
        d->timeout_id = g_timeout_add_seconds (item->timeout,
                                               _daemon_show_next_item,
                                               d);
    }

    DEBUG ("end");

    return FALSE;
}

void
_daemon_browser_stop_session (WkDaemon *d)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    DEBUG ("session stop");

    wk_browser_clear_session (browser, TRUE);
    d->status = WK_DAEMON_STATUS_INVALID;
}

void
_daemon_browser_start_session (WkDaemon *d)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));

    if (d->status == WK_DAEMON_STATUS_BROWSER)
        return;

    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);

    DEBUG ("session start");

    if (d->settings->slideshow_hide_pointer)
        stop_unclutter ();

    wk_browser_start_session (browser);
    d->status = WK_DAEMON_STATUS_BROWSER;
}

void
_daemon_handle_post_session (WkDaemon *d)
{
    switch (d->settings->after_web_session)
    {
        case AFTER_WEB_SESSION_SLIDESHOW:
            wk_daemon_start_slideshow (d);
            break;
        case AFTER_WEB_SESSION_SESSION:
            if (d->settings->with_timer)
                _daemon_browser_show_uri (d, d->settings->default_url);
            else
                _daemon_browser_start_session (d);
            break;
        default:
            _daemon_browser_show_uri (d, d->settings->default_url);
            break;
    }
}

gboolean
_daemon_post_session_timeout (gpointer data)
{
    WkDaemon *d = (WkDaemon *) data;
    d->status = WK_DAEMON_STATUS_INVALID;

    _daemon_handle_post_session ((WkDaemon *) data);

    return FALSE;
}

static void
_daemon_get_playlist_data (WkDaemon *d)
{
    DEBUG ("start");

    wk_pl_reset (d->playlist);

    g_free (d->playlist_src);

    /* start loading the slideshow infos if any */
    DEBUG ("slideshow_file: %s", d->settings->slideshow_file);
    if (d->settings->slideshow_file[0] == '/')
        g_file_get_contents (d->settings->slideshow_file,
                             &d->playlist_src, NULL, NULL);
    else if (strncmp (d->settings->slideshow_file, "file://", 7) == 0)
        g_file_get_contents (d->settings->slideshow_file + 7,
                             &d->playlist_src, NULL, NULL);
    else
        d->playlist_src = get_remote_data (d->settings->slideshow_file);

    /* if we couldn't get the data, try to use the cache */
    if (!d->playlist_src)
    {
        DEBUG ("getting XML data from cache");
        if (!g_file_get_contents (d->playlist_file, &d->playlist_src, NULL, NULL))
            d->playlist_src = NULL;
    }

    /* if no cache is usable, bail */
    if (!d->playlist_src)
        return;

    /* otherwise parse the playlist and (re)get the data associated */
    if (! wk_pl_set_from_xml (d->playlist, d->playlist_src))
    {
        if (! wk_pl_set_from_list (d->playlist, d->playlist_src))
        {
            /* failed to parse the playlist, same result as if it didn't exist */
            g_free (d->playlist_src);
            d->playlist_src = NULL;
            return;
        }
    }

    DEBUG ("new playlist will be used");

    /* cache the playlist */
    g_file_set_contents (d->playlist_file, d->playlist_src, -1, NULL);

    /* get the cache for data */
    wk_pl_set_cache (d->playlist, d->data_path, d->etag_path);

    /* select the best fitting loop from the playlist infos */
    _daemon_select_playlist_loop (d);

#ifdef WITH_DEBUG
    wk_pl_print (d->playlist);
#endif
}

static void
_daemon_select_playlist_loop (WkDaemon *d)
{
    wk_pl_select (d->playlist);
    d->loop = d->playlist->selected_loop;
}

static gboolean
_on_playlist_timeout (WkDaemon *d)
{
    DEBUG ("Playlist timeout");

    /* invalidate the current loop and item to handle possible changes */
    /* FIXME: what if nothing changed? we might restart a still valid loop */
    d->loop = NULL;
    d->current = NULL;

    _daemon_get_playlist_data (d);

    return TRUE;
}

WkDaemon *
wk_daemon_get ()
{
    static WkDaemon *d = NULL;
    const gchar *home;

    if (d)
        return d;

    d = g_new (WkDaemon, 1);
    d->status = WK_DAEMON_STATUS_INVALID;
    d->current = NULL;
    d->settings = GET_SETTINGS ();
    d->playlist = wk_pl_new ();
    d->loop = NULL;

    home = getenv ("HOME");
    d->data_path = g_strdup_printf ("%s/.linutop-kiosk/data/", home);
    g_mkdir_with_parents (d->data_path, 0755);
    d->etag_path = g_strdup_printf ("%s/.linutop-kiosk/etag/", home);
    g_mkdir_with_parents (d->etag_path, 0755);
    d->playlist_file = g_strdup_printf ("%s/.linutop-kiosk/playlist", home);

    d->playlist_src = NULL;

    return d;
}

void
wk_daemon_start (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_INVALID)
        return;

    if (d->settings->mode == MODE_KIOSK)
    {
        if (d->settings->prestart_file &&
               (g_access (d->settings->prestart_file, X_OK) == 0))
        {
            gint status;

            if (g_spawn_command_line_sync (d->settings->prestart_file,
                                       NULL, NULL, &status, NULL)
                && WIFEXITED (status)
                && (WEXITSTATUS (status) == 0))
            {
                _daemon_browser_start_session (d);
            }
        }
        else if (d->settings->start_with_session)
            _daemon_browser_start_session (d);
    }

    /* in display or kiosk+display mode only */
    if ((d->settings->mode != MODE_KIOSK)
        || (d->settings->after_web_session == AFTER_WEB_SESSION_SLIDESHOW))
    {
        _daemon_get_playlist_data (d);
    }

    /* stop here if we've started something */
    if (d->status != WK_DAEMON_STATUS_INVALID)
        return;

    if (d->loop)
    {
        DEBUG ("starting slideshow");
        wk_daemon_start_slideshow (d);
    }
    else
    {
        DEBUG ("displaying default URL");
        _daemon_browser_show_uri (d, d->settings->default_url);
    }
}

void
wk_daemon_start_slideshow (WkDaemon *d)
{
    DEBUG ("start");

    if (d->status == WK_DAEMON_STATUS_SLIDESHOW)
        return;
    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);
    if (d->status == WK_DAEMON_STATUS_BROWSER)
        _daemon_browser_stop_session (d);

    if (d->settings->slideshow_hide_pointer)
        start_unclutter ();

    d->status = WK_DAEMON_STATUS_SLIDESHOW;
    d->current = NULL;

    if (d->settings->slideshow_refresh_after_minutes)
    {
        DEBUG ("refreshing playlist every %d minutes",
               d->settings->slideshow_refresh_after_minutes);
        d->playlist_timeout_id = g_timeout_add_seconds (
                                    d->settings->slideshow_refresh_after_minutes*60,
                                    (GSourceFunc) _on_playlist_timeout, d);
    }

    wk_event_set_event_filter ();

    _daemon_show_next_item (d);

    DEBUG ("end");
}

void
wk_daemon_handle_start_session (WkDaemon *d)
{
    GSource *source;

    DEBUG ("start");

    if (d->settings->mode != MODE_KIOSK)
        return;

    if (d->status == WK_DAEMON_STATUS_BROWSER)
        return;

    if (d->status == WK_DAEMON_STATUS_IDLE)
        g_spawn_command_line_async ("xset dpms force on", NULL);

    d->status = WK_DAEMON_STATUS_INVALID;

    if (d->timeout_id
        && (source = g_main_context_find_source_by_id (
                    NULL, d->timeout_id)) != NULL)
    {
        g_source_destroy (source);
    }

    if (d->playlist_timeout_id
        && (source = g_main_context_find_source_by_id (
                    NULL, d->playlist_timeout_id)) != NULL)
    {
        g_source_destroy (source);
    }

    _daemon_browser_start_session (d);

    DEBUG ("end");
}

void
wk_daemon_handle_stop_session (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_BROWSER)
        return;

    wk_browser_clear_session (WK_BROWSER (wk_browser_get (NULL)), TRUE);
    d->status = WK_DAEMON_STATUS_INVALID;

    if (d->settings->post_session_uri)
    {
        _daemon_browser_show_uri (d, d->settings->post_session_uri);
        g_timeout_add_seconds (d->settings->post_session_uri_time,
                               _daemon_post_session_timeout,
                               d);
    }
    else
        _daemon_handle_post_session (d);
}

void
wk_daemon_handle_video_end (WkDaemon *d)
{
    if (d->status != WK_DAEMON_STATUS_SLIDESHOW)
        return;

    _daemon_show_next_item (d);
}
